/*
 * PrimeBubblesSonic: An A/V thing based on prime factorisation.
 *
 * Requires SuperCollider 3 to be running, with the synthdefs in the
 * SynthDefs tab installed, plus the p5_sc library:
 *   <http://www.erase.net/projects/p5_sc/>
 *
 * Copyright (c) Daniel Jones 2009.
 *   <http://www.erase.net/>
 * 
 * This program can be freely distributed and modified under the
 * terms of the GNU General Public License version 2.
 *   <http://www.gnu.org/copyleft/gpl.html>
 *
 *----------------------------------------------------------------------*/
 
import processing.opengl.*;
import javax.media.opengl.*;
import com.sun.opengl.util.texture.*;

import supercollider.*;
import oscP5.*;

import processing.video.*;
MovieMaker mm;

PGraphicsOpenGL pgl;
GL gl;

/*----------------------------------------------------------------------*
 * How fast do you want it?
 *----------------------------------------------------------------------*/
final int FRAME_RATE = 25;
final int FRAMES_PER_NUMBER = 3;

/*----------------------------------------------------------------------*
 * Fundamental frequency that partials are based upon.
 *----------------------------------------------------------------------*/
final float FREQ_BASE = 65.0;

/*----------------------------------------------------------------------*
 * The triad of fundamental frequencies, cycled through every 64 notes.
 *----------------------------------------------------------------------*/
final float [] FREQ_RATIOS = new float [] { 1/1.0, 6/5.0, 11/12.0 };

/*----------------------------------------------------------------------*
 * Drawing options.
 *----------------------------------------------------------------------*/
final boolean DRAW_TEXT = true;
final float   DRAW_FADE = 0.5;
final boolean DRAW_CAPTURE_FRAMES = false;
final boolean DRAW_RECORD = false;
final boolean DRAW_MONOCHROME = false;

final boolean START_PAUSED = false;

/*----------------------------------------------------------------------*
 * Use random sequence rather than ordered integers?
 *----------------------------------------------------------------------*/
final boolean COUNT_RANDOMIZE = false;

//////////////////////////////////////////////////////////////////////////

Product p;
int number = 0;
int frame_count = 0;

int [] indices;
float [] hues;
float [] freqs;
boolean paused = START_PAUSED;

int freq_ratio_count = 0;

PImage background;
  
Synth harmonics[];

PFont font;
float c;

Texture circle;

void setup()
{
  size(1440, 900, OPENGL);
//  hint(DISABLE_OPENGL_2X_SMOOTH);  
  colorMode(HSB, 1.0);
  smooth();
  frameRate(FRAME_RATE);
  initGL();
  
  indices = new int[40000];
  populate_indices();
  
  hues = new float[40000];
  populate_hues();
  
  freqs = new float[40000];
  populate_freqs();
  
  font = loadFont("HelveticaNeue-Bold-72.vlw"); 
  textFont(font, 48);
  textAlign(CENTER);
  
  background(0.7, 0.8, 0.15);

  /*----------------------------------------------------------------------*
   * Circle images are drawn from a texture for nice glowing edges.
   *----------------------------------------------------------------------*/
  try { circle = TextureIO.newTexture(new File(dataPath("circle-glow.png")), true); }
  catch(Exception e) { println(e); }  
  
  /*----------------------------------------------------------------------*
   * Create all harmonic synths at the outset. These are then simply
   * turned on and off when needed. 
   *----------------------------------------------------------------------*/
  harmonics = new Synth[400];
  for (int i = 0; i < harmonics.length; i++)
  {
    harmonics[i] = new Synth("inst_harmonic");
    harmonics[i].set("freq", FREQ_BASE * (i + 1));
    harmonics[i].set("delay", 0.2);
    harmonics[i].set("on", 0);
    harmonics[i].set("amp", 0.05);      
    harmonics[i].create();
  }
  
  background = loadImage("background.png");
  image(background, 0, 0);
  
  if (DRAW_RECORD)
    mm = new MovieMaker(this, width, height, "drawing.mov",
                       25, MovieMaker.JPEG, MovieMaker.HIGH);
}

/*----------------------------------------------------------------------*
 * This is where the action happens.
 *  - increment the number counter, and update the base freq
 *  - clear the frame with a background image
 *  - stop the previous frame's audio
 *  - generate and draw the new Product object
 *  - start the new frame's audio
 *  - record the current frame, if necessary
 *----------------------------------------------------------------------*/
void draw()
{
  if (paused)
  {
    if (DRAW_RECORD)
      mm.addFrame();  // Add window's pixels to movie    
    return;
  }
    
  if ((frame_count % (64 * FRAMES_PER_NUMBER)) == 0)
    renew_freqs();

  if ((frame_count % FRAMES_PER_NUMBER) == 0)
    if (COUNT_RANDOMIZE)
      number = (int) random(0, 1500);
    else
      number++;

  // Unused now we have a background image.    
//  background(0.7, 0.6, 0.2);
//  fill(0.7, 0.8, 0.15, DRAW_FADE);
//  noStroke();
//  rect(0, 0, width, height);

  tint(1, DRAW_FADE);
  image(background, 0, 0);
  
  initGLdraw();  
  
  if ((frame_count % FRAMES_PER_NUMBER) == 0)  
    if (p != null) p.soundoff();
    
  p = new Product(number, width/2, height/2);
  p.scale = 25.5;

//  p = new Product(number, 25 + 49.5 * (number % 29), 25 + 50 * ((int) number / 29));
//  p.scale = 1.5;
   
  
  p.draw();
  if ((frame_count % FRAMES_PER_NUMBER) == 0)
    p.sound();
  
  if (DRAW_CAPTURE_FRAMES && ((frame_count % FRAMES_PER_NUMBER) == 1))
  {
    String filename = "frames/" + number + ".png";
    File frame = new File(sketchPath(filename));
    println("frame is " + sketchPath(filename));
    if (!frame.exists())
      saveFrame("frames/" + number + ".png");
  }
  
  frame_count++;
  
  if (DRAW_RECORD)
    mm.addFrame();  // Add window's pixels to movie

}

void _draw()
{
  int number = frameCount + 1;
  
  if (p != null) p.soundoff();
  p = new Product(number, 25 + 20 * (frameCount % 60), 25 + 20 * ((int) frameCount / 60));
  p.scale = 0.8;
  p.draw();  
  p.sound();
}

/*----------------------------------------------------------------------*
 * Triggered when esc is pressed. Kill all sc synths, and
 * finish movie rendering.
 *----------------------------------------------------------------------*/
void stop()
{
  for (int i = 0; i < harmonics.length; i++)
    harmonics[i].free();
  
  if (DRAW_RECORD)
    mm.finish();
}

/*----------------------------------------------------------------------*
 * Populate prime index array.
 *----------------------------------------------------------------------*/
void populate_indices()
{
  int index = 0;
  int prime = 1;
  
  for (int i = 0; i < 2000; i++)
  {
    prime++;
    while (!isPrime(prime))
      prime++;
    
    indices[prime] = index++;
  }
}

/*----------------------------------------------------------------------*
 * Populate hue array corresponding to prime indices.
 *----------------------------------------------------------------------*/
void populate_hues()
{
  float hue = 0f;
  
  int count = 1;
  for (int i = 0; i < 2000; i++)
  {
    count++;
    while (!isPrime(count))
      count++;
    
//    println("c " + count + ": " + hue);
    hues[count] = hue;
    hue = (hue + 0.174);
    hue -= (int) hue;
  }
}

/*----------------------------------------------------------------------*
 * Populate partial frequency array.
 *----------------------------------------------------------------------*/
void populate_freqs()
{
  float freq = FREQ_BASE;

  int count = 1;
  for (int i = 0; i < 200; i++)
  {
    count++;
    while (!isPrime(count))
      count++;
    
    freqs[count] = freq;
    freq += FREQ_BASE;
  }  
}

/*----------------------------------------------------------------------*
 * Update fundamental frequency and harmonics when we hit a
 * new freq in the base triad.
 *----------------------------------------------------------------------*/
void renew_freqs()
{
  freq_ratio_count = (freq_ratio_count + 1) % FREQ_RATIOS.length;
  float freq = FREQ_BASE * FREQ_RATIOS[freq_ratio_count];

  for (int i = 0; i < 200; i++)
    harmonics[i].set("freq", freq * (i + 1));
}

/*----------------------------------------------------------------------*
 * Not going to win any prizes for efficiency - brute force
 * test for primeness.
 *----------------------------------------------------------------------*/
boolean isPrime(int n)
{
  for (int i = 2; i <= sqrt(n); i++)
    if (n % i == 0) return false;

  return true;
}

/*----------------------------------------------------------------------*
 * Pause if we hit space.
 *----------------------------------------------------------------------*/
void keyPressed()
{
  if (key == ' ')
    paused = !paused;
}    
